// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2023 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <algorithm>
#include <sstream>
#include <iostream>
#include "core/log.h"
#include "core/types.h"
#include "core/guid.h"
#include "core/data_scheduler.h"

using namespace SEAScope;

/**
 *
 */
DataScheduler::DataScheduler()
{
};

/**
 *
 */
DataScheduler::~DataScheduler()
{
};

/**
 *
 */
static void buildJobId(const std::string& action, const GranuleId& granuleId, const std::string& uri, const std::set<std::string>& fields, std::string& result)
{
    std::ostringstream buffer {""};
    buffer << action << "(" << granuleId.first << "/" << granuleId.second << "_";
    for
        (const auto& field : fields)
    {
        buffer << field << "+";
    }
    buffer << "_" << uri << ")";
    result = buffer.str();
}

/**
 *
 */
static bool isJobAlreadyQueued(const std::list<Job>& jobs, const std::string& jobId)
{
    auto found = std::find_if(jobs.begin(), jobs.end(),
                              [&jobId] (const Job& pendingJob) { return jobId == pendingJob.id; });
    return jobs.end() != found;
}

/**
 *
 */
void DataScheduler::removeRelatedJobs(const DataLoadingJob& j)
{
	std::ostringstream buffer {""};	//
    std::string toCancel {""};  //
    std::string loadToCancel {""};  //

    // Build pattern for the beginning of download job ids for the same
    // granule and fields
    buffer << "downloadData(" << j.granule.first << "/" << j.granule.second << "_";
    for
        (const auto& field : j.fields)
    {
        buffer << field << "+";
    }

    toCancel = buffer.str();
    std::size_t toCancelSize = toCancel.size();

    // Build id for load job related to the same granule and fields
    buildJobId("loadData", j.granule, "", j.fields, loadToCancel);

    auto it = std::remove_if(this->_jobs.begin(),
                             this->_jobs.end(),
                             [&toCancel, &toCancelSize, &loadToCancel] (const Job& pendingJob) {
        return (  (loadToCancel == pendingJob.id)
               || (  (toCancelSize < pendingJob.id.size())
                  && 0 == pendingJob.id.compare(0, toCancelSize, toCancel))) ;
    });
    this->_jobs.erase(it, this->_jobs.end());
}

/**
 *
 */
bool DataScheduler::addJob(Job& job, std::string& jobId)
{
	bool ok {false};	//
    std::string toCancel {""};  //
	DataLoadingJob* j {nullptr};	//

	j = dynamic_cast<DataLoadingJob*>(&job);
	if
		(nullptr == j)
	{
		Log::log(LogLevel::Error, "DataScheduler", "The job is not a DataLoadingJob");
		return false;
	}

	if
		("downloadData" == job.name)
	{
        buildJobId("downloadData", j->granule, j->uri, j->fields, job.id);

        ok = isJobAlreadyQueued(this->_jobs, job.id);
        if
            (true == ok)
        {
            // No need to queue the job again if a duplicate already exists
            job.status = JobStatus::CANCELLED;
            return true;
        }

        this->removeRelatedJobs(*j);
	}
	else if
        ("loadData" == job.name)
	{
        buildJobId("loadData", j->granule, "", j->fields, job.id);

        this->removeRelatedJobs(*j);
    }
    else
    {
		ok = GUID::generate(job.id);
		if
			(false == ok)
		{
			Log::log( LogLevel::Error, "DataScheduler"
					, "Could not generate GUID for new job");
			return false;
		}
	}

	Log::log( LogLevel::Debug, "Datascheduler"
			, "Add data job: {} => {}", job.name, job.id);

    ok = isJobAlreadyQueued(this->_jobs, job.id);
    if
        (true == ok)
    {
        // No need to queue the job again if a duplicate already exists
        job.status = JobStatus::CANCELLED;
        return true;
    }

    // Pass the generated identifier to the caller
    jobId = job.id;

	job.status = JobStatus::PENDING;
	this->_jobs.push_back(job);

	return true;
};

/**
 *
 */
bool DataScheduler::removeJob(const std::string& jobId)
{
	auto it = std::remove_if( this->_jobs.begin(),
				              this->_jobs.end(),
				              [&jobId] (const Job& j) -> bool { return j.id == jobId; });
    this->_jobs.erase(it, this->_jobs.end());

	return true;
};

/**
 *
 */
bool DataScheduler::selectJob(Job& job)
{
	if
		(this->_jobs.empty())
	{
		return false;
	}

    job = std::move(*this->_jobs.begin());
    this->_jobs.erase(this->_jobs.begin());

	return true;
};

/**
 *
 */
bool DataScheduler::updateJobStatus(const std::string& jobId, const JobStatus& status)
{
	auto found = std::find_if( this->_jobs.begin()
							 , this->_jobs.end()
							 , [jobId] (const Job& j) { return j.id == jobId; });

	if
		(  this->_jobs.end() == found
		|| found->status == status)
	{
		return false;
	}

	found->status = status;
	return true;
};

/**
 *
 */
bool DataScheduler::listJobs(std::list<Job>& result)
{
	result = this->_jobs;
	return true;
};

/**
 *
 */
bool DataScheduler::listJobs(std::list<std::string>& result)
{
	for
		(const auto& job : this->_jobs)
	{
		result.push_back(job.id);
	}

	return true;
};

/**
 *
 */
std::size_t DataScheduler::countPendingJobs() const
{
	std::int64_t result;	//

	result = std::count_if( this->_jobs.begin()
						  , this->_jobs.end()
						  , [](const Job& j) { return JobStatus::PENDING == j.status; });
	return static_cast<std::size_t>(result);
};
