// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2023 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include "core/granule_metadata.h"
#include "core/log.h"
#include "core/structs.h"

#include <iostream>

using namespace SEAScope;

constexpr std::uint32_t SourceIdIndex = 0;
constexpr std::uint32_t CollectionIdIndex = SourceIdIndex + 4;
constexpr std::uint32_t GranuleIdIndex = CollectionIdIndex + 4;
constexpr std::uint32_t DataModelIndex = GranuleIdIndex + 8;
constexpr std::uint32_t AvailableAttributesIndex = DataModelIndex + 1;
constexpr std::uint32_t TimeCoverageStartIndex = AvailableAttributesIndex + 4;
constexpr std::uint32_t TimeCoverageEndIndex = TimeCoverageStartIndex + 8;
constexpr std::uint32_t LatMinIndex = TimeCoverageEndIndex + 8;
constexpr std::uint32_t LatMaxIndex = LatMinIndex + 4;
constexpr std::uint32_t LonMinIndex = LatMaxIndex + 4;
constexpr std::uint32_t LonMaxIndex = LonMinIndex + 4;
constexpr std::uint32_t StringsOffsetIndex = LonMaxIndex + 4;
constexpr std::uint32_t DataIdIndex = StringsOffsetIndex + 4;
constexpr std::uint32_t TitleIndex = DataIdIndex + 4;
constexpr std::uint32_t InstitutionIndex = TitleIndex + 4;
constexpr std::uint32_t CommentIndex = InstitutionIndex + 4;
constexpr std::uint32_t FileIdIndex = CommentIndex + 4;
constexpr std::uint32_t ProductVersionIndex = FileIdIndex + 4;
constexpr std::uint32_t CreatorEmailIndex = ProductVersionIndex + 4;
constexpr std::uint32_t StationIdIndex = CreatorEmailIndex + 4;
constexpr std::uint32_t PlatformIndex = StationIdIndex + 4;
constexpr std::uint32_t SensorIndex = PlatformIndex + 4;
constexpr std::uint32_t GcpsOffsetIndex = SensorIndex + 4;
constexpr std::uint32_t ViewsOffsetIndex = GcpsOffsetIndex + 4;

/**
 *
 */
GranuleMetadata::GranuleMetadata()
{
}


/**
 * Assumptions:
 * - the cumulative length of strings will not exceed the max value for
 *   std::uint32_t
 * - the cumulative length of gcps will not exceed the max value for
 *   std::uint32_t
 */
void GranuleMetadata::build(std::uint32_t sourceId,
                                std::uint32_t collectionId,
                                std::uint64_t granuleId,
                                DataModel dataModel,
                                timestamp_t timeCoverageStart,
                                timestamp_t timeCoverageEnd,
                                bool hasLatMin,
                                float latMin,
                                bool hasLatMax,
                                float latMax,
                                bool hasLonMin,
                                float lonMin,
                                bool hasLonMax,
                                float lonMax,
                                const std::string& dataId,
                                const std::string& title,
                                const std::string& institution,
                                const std::string& comment,
                                const std::string& fileId,
                                const std::string& productVersion,
                                const std::string& creatorEmail,
                                const std::string& stationId,
                                const std::string& platform,
                                const std::string& sensor,
                                const IDFDescriptor& descriptor)
{
    std::int32_t int32Buffer {0};

    this->_blob.clear();

    // Reserve memory for all the fields that are not owned by the
    // IDFDescriptor (it could be included later on if optimization is needed)
    this->_blob.reserve( 3 * sizeof(std::uint64_t) /*granuleId + timeCoverageStart + timeCoverageEnd*/
                       + 15 * sizeof(std::uint32_t) /*sourceId, collectionId + stringsOffset + dataIdEnd + titleEnd + institutionEnd + commentEnd + fileIdEnd + productVersionEnd + creatorEmailEnd + stationIdEnd + platformEnd + sensorEnd + gcpsOffset* + availableAttributes*/
                       + 1 * sizeof(std::uint8_t) /*dataModel*/
                       + 4 * sizeof(float) /*latMin + latMax + lonMin + lonMax*/
                       + dataId.length()
                       + title.length()
                       + institution.length()
                       + comment.length()
                       + fileId.length()
                       + productVersion.length()
                       + creatorEmail.length()
                       + stationId.length()
                       + platform.length()
                       + sensor.length());


    // Granule ID - source
    this->_blob.emplace_back((sourceId >> 0) & 0xFF);
    this->_blob.emplace_back((sourceId >> 8) & 0xFF);
    this->_blob.emplace_back((sourceId >> 16) & 0xFF);
    this->_blob.emplace_back((sourceId >> 24) & 0xFF);

    // Collection ID
    this->_blob.emplace_back((collectionId >> 0) & 0xFF);
    this->_blob.emplace_back((collectionId >> 8) & 0xFF);
    this->_blob.emplace_back((collectionId >> 16) & 0xFF);
    this->_blob.emplace_back((collectionId >> 24) & 0xFF);

    // Granule ID - granule
    this->_blob.emplace_back((granuleId >> 0) & 0xFF);
    this->_blob.emplace_back((granuleId >> 8) & 0xFF);
    this->_blob.emplace_back((granuleId >> 16) & 0xFF);
    this->_blob.emplace_back((granuleId >> 24) & 0xFF);
    this->_blob.emplace_back((granuleId >> 32) & 0xFF);
    this->_blob.emplace_back((granuleId >> 40) & 0xFF);
    this->_blob.emplace_back((granuleId >> 48) & 0xFF);
    this->_blob.emplace_back((granuleId >> 56) & 0xFF);

    // Data model
    this->_blob.emplace_back(static_cast<std::uint8_t>(dataModel));

    // Attributes availability
    std::uint32_t availableAttributes {0u};
    if
        (0 < title.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::TITLE));
    }
    if
        (0 < institution.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::INSTITUTION));
    }
    if
        (0 < comment.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::COMMENT));
    }
    if
        (0 < fileId.length())
    {
        availableAttributes = availableAttributes | ( 1 << static_cast<std::size_t>(IDFAttributes::FILE_ID));
    }
    if
        (0 < productVersion.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::PRODUCT_VERSION));
    }
    if
        (true == hasLatMin)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LAT_MIN));
    }
    if
        (true == hasLatMax)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LAT_MAX));
    }
    if
        (true == hasLonMin)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LON_MIN));
    }
    if
        (true == hasLonMax)
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::LON_MAX));
    }
    if
        (0 < creatorEmail.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::CREATOR_EMAIL));
    }
    if
        (0 < stationId.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::STATION_ID));
    }
    if
        (0 < platform.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::PLATFORM));
    }
    if
        (0 < sensor.length())
    {
        availableAttributes = availableAttributes | (1 << static_cast<std::size_t>(IDFAttributes::SENSOR));
    }
    this->_blob.emplace_back((availableAttributes >> 0) & 0xFF);
    this->_blob.emplace_back((availableAttributes >> 8) & 0xFF);
    this->_blob.emplace_back((availableAttributes >> 16) & 0xFF);
    this->_blob.emplace_back((availableAttributes >> 24) & 0xFF);

    // Time coverage start
    this->_blob.emplace_back((timeCoverageStart >> 0) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 8) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 16) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 24) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 32) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 40) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 48) & 0xFF);
    this->_blob.emplace_back((timeCoverageStart >> 56) & 0xFF);

    // Time coverage end
    this->_blob.emplace_back((timeCoverageEnd >> 0) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 8) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 16) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 24) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 32) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 40) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 48) & 0xFF);
    this->_blob.emplace_back((timeCoverageEnd >> 56) & 0xFF);

    // Lat min
    memcpy(&int32Buffer, &latMin, 4);
    this->_blob.emplace_back((int32Buffer >> 0) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 8) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 16) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 24) & 0xFF);

    // Lat max
    memcpy(&int32Buffer, &latMax, 4);
    this->_blob.emplace_back((int32Buffer >> 0) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 8) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 16) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 24) & 0xFF);

    // Lon min
    memcpy(&int32Buffer, &lonMin, 4);
    this->_blob.emplace_back((int32Buffer >> 0) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 8) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 16) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 24) & 0xFF);

    // Lon Max
    memcpy(&int32Buffer, &lonMax, 4);
    this->_blob.emplace_back((int32Buffer >> 0) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 8) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 16) & 0xFF);
    this->_blob.emplace_back((int32Buffer >> 24) & 0xFF);

    // Strings offset
    std::uint32_t stringsOffset {GcpsOffsetIndex + 4u};
    this->_blob.emplace_back((stringsOffset >> 0) & 0xFF);
    this->_blob.emplace_back((stringsOffset >> 8) & 0xFF);
    this->_blob.emplace_back((stringsOffset >> 16) & 0xFF);
    this->_blob.emplace_back((stringsOffset >> 24) & 0xFF);

    // Data ID
    std::uint32_t dataIdEnd {static_cast<std::uint32_t>(dataId.length())};
    this->_blob.emplace_back((dataIdEnd >> 0) & 0xFF);
    this->_blob.emplace_back((dataIdEnd >> 8) & 0xFF);
    this->_blob.emplace_back((dataIdEnd >> 16) & 0xFF);
    this->_blob.emplace_back((dataIdEnd >> 24) & 0xFF);

    // Title
    std::uint32_t titleEnd {dataIdEnd + static_cast<std::uint32_t>(title.length())};
    this->_blob.emplace_back((titleEnd >> 0) & 0xFF);
    this->_blob.emplace_back((titleEnd >> 8) & 0xFF);
    this->_blob.emplace_back((titleEnd >> 16) & 0xFF);
    this->_blob.emplace_back((titleEnd >> 24) & 0xFF);

    // Institution
    std::uint32_t institutionEnd {titleEnd + static_cast<std::uint32_t>(institution.length())};
    this->_blob.emplace_back((institutionEnd >> 0) & 0xFF);
    this->_blob.emplace_back((institutionEnd >> 8) & 0xFF);
    this->_blob.emplace_back((institutionEnd >> 16) & 0xFF);
    this->_blob.emplace_back((institutionEnd >> 24) & 0xFF);

    // Comment
    std::uint32_t commentEnd {institutionEnd + static_cast<std::uint32_t>(comment.length())};
    this->_blob.emplace_back((commentEnd >> 0) & 0xFF);
    this->_blob.emplace_back((commentEnd >> 8) & 0xFF);
    this->_blob.emplace_back((commentEnd >> 16) & 0xFF);
    this->_blob.emplace_back((commentEnd >> 24) & 0xFF);

    // File ID
    std::uint32_t fileIdEnd {commentEnd + static_cast<std::uint32_t>(fileId.length())};
    this->_blob.emplace_back((fileIdEnd >> 0) & 0xFF);
    this->_blob.emplace_back((fileIdEnd >> 8) & 0xFF);
    this->_blob.emplace_back((fileIdEnd >> 16) & 0xFF);
    this->_blob.emplace_back((fileIdEnd >> 24) & 0xFF);

    // Product version
    std::uint32_t productVersionEnd {fileIdEnd + static_cast<std::uint32_t>(productVersion.length())};
    this->_blob.emplace_back((productVersionEnd >> 0) & 0xFF);
    this->_blob.emplace_back((productVersionEnd >> 8) & 0xFF);
    this->_blob.emplace_back((productVersionEnd >> 16) & 0xFF);
    this->_blob.emplace_back((productVersionEnd >> 24) & 0xFF);

    // Creator email
    std::uint32_t creatorEmailEnd {productVersionEnd + static_cast<std::uint32_t>(creatorEmail.length())};
    this->_blob.emplace_back((creatorEmailEnd >> 0) & 0xFF);
    this->_blob.emplace_back((creatorEmailEnd >> 8) & 0xFF);
    this->_blob.emplace_back((creatorEmailEnd >> 16) & 0xFF);
    this->_blob.emplace_back((creatorEmailEnd >> 24) & 0xFF);

    // Station ID
    std::uint32_t stationIdEnd {creatorEmailEnd + static_cast<std::uint32_t>(stationId.length())};
    this->_blob.emplace_back((stationIdEnd >> 0) & 0xFF);
    this->_blob.emplace_back((stationIdEnd >> 8) & 0xFF);
    this->_blob.emplace_back((stationIdEnd >> 16) & 0xFF);
    this->_blob.emplace_back((stationIdEnd >> 24) & 0xFF);

    // Platform
    std::uint32_t platformEnd {stationIdEnd + static_cast<std::uint32_t>(platform.length())};
    this->_blob.emplace_back((platformEnd >> 0) & 0xFF);
    this->_blob.emplace_back((platformEnd >> 8) & 0xFF);
    this->_blob.emplace_back((platformEnd >> 16) & 0xFF);
    this->_blob.emplace_back((platformEnd >> 24) & 0xFF);

    // Sensor
    std::uint32_t sensorEnd {platformEnd + static_cast<std::uint32_t>(sensor.length())};
    this->_blob.emplace_back((sensorEnd >> 0) & 0xFF);
    this->_blob.emplace_back((sensorEnd >> 8) & 0xFF);
    this->_blob.emplace_back((sensorEnd >> 16) & 0xFF);
    this->_blob.emplace_back((sensorEnd >> 24) & 0xFF);

    // GCPs offset
    std::uint32_t gcpsOffset {stringsOffset + sensorEnd};
    this->_blob.emplace_back((gcpsOffset >> 0) & 0xFF);
    this->_blob.emplace_back((gcpsOffset >> 8) & 0xFF);
    this->_blob.emplace_back((gcpsOffset >> 16) & 0xFF);
    this->_blob.emplace_back((gcpsOffset >> 24) & 0xFF);

    // Serialize strings
    if
        (0 < dataId.length())
    {
        this->_blob.insert(this->_blob.end(), dataId.begin(), dataId.end());
    }

    if
        (0 < title.length())
    {
        this->_blob.insert(this->_blob.end(), title.begin(), title.end());
    }

    if
        (0 < institution.length())
    {
        this->_blob.insert(this->_blob.end(), institution.begin(), institution.end());
    }

    if
        (0 < comment.length())
    {
        this->_blob.insert(this->_blob.end(), comment.begin(), comment.end());
    }

    if
        (0 < fileId.length())
    {
        this->_blob.insert(this->_blob.end(), fileId.begin(), fileId.end());
    }

    if
        (0 < productVersion.length())
    {
        this->_blob.insert(this->_blob.end(), productVersion.begin(), productVersion.end());
    }

    if
        (0 < creatorEmail.length())
    {
        this->_blob.insert(this->_blob.end(), creatorEmail.begin(), creatorEmail.end());
    }

    if
        (0 < stationId.length())
    {
        this->_blob.insert(this->_blob.end(), stationId.begin(), stationId.end());
    }

    if
        (0 < platform.length())
    {
        this->_blob.insert(this->_blob.end(), platform.begin(), platform.end());
    }

    if
        (0 < sensor.length())
    {
        this->_blob.insert(this->_blob.end(), sensor.begin(), sensor.end());
    }

    // IDFDescriptor
    this->addIDFDescriptor(descriptor);
}

/**
 *
 */
void GranuleMetadata::addIDFDescriptor(const IDFDescriptor& descriptor)
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    constexpr std::uint32_t gcpSize = sizeof(std::uint64_t) + 2 * (sizeof(std::uint32_t) + sizeof(float));
    const std::uint32_t uriSize {static_cast<std::uint32_t>(descriptor.uri.length())};
    const std::uint32_t originalSize {static_cast<std::uint32_t>(this->_blob.size())};

    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get current gcpsOffset
    std::uint32_t gcpsOffset {0u};
    gcpsOffset = ( (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 3]) << 24));

    // IDFDescriptors must be stored in ascending order of spacing (i.e.
    // the IDF with the best resolution first)
    float existingSpacing {0.f};
    std::int32_t int32Buffer {0};
    std::uint32_t viewOffset = ViewsOffsetIndex;
    while
        (viewOffset < stringsOffset)
    {
        int32Buffer = ( (static_cast<std::int32_t>(this->_blob[viewOffset + 8]) << 0)
                      | (static_cast<std::int32_t>(this->_blob[viewOffset + 9]) << 8)
                      | (static_cast<std::int32_t>(this->_blob[viewOffset + 10]) << 16)
                      | (static_cast<std::int32_t>(this->_blob[viewOffset + 11]) << 24));
        memcpy(&existingSpacing, &int32Buffer, 4);
        if
            (existingSpacing > descriptor.resolution)
        {
            break;
        }
        viewOffset = viewOffset + descriptorSize;
        //std::cout << "viewOffset = " << std::to_string(viewOffset) << " < stringsOffset = " << std::to_string(stringsOffset) << std::endl;
    }

    // xArity
    this->_blob.insert(this->_blob.begin() + viewOffset,
                       (descriptor.xArity >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 1,
                       (descriptor.xArity >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 2,
                       (descriptor.xArity >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 3,
                       (descriptor.xArity >> 24) & 0xFF);

    // yArity
    this->_blob.insert(this->_blob.begin() + viewOffset + 4,
                       (descriptor.yArity >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 5,
                       (descriptor.yArity >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 6,
                       (descriptor.yArity >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 7,
                       (descriptor.yArity >> 24) & 0xFF);

    // spacing in meters
    memcpy(&int32Buffer, &(descriptor.resolution), 4);
    this->_blob.insert(this->_blob.begin() + viewOffset + 8,
                       (int32Buffer >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 9,
                       (int32Buffer >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 10,
                       (int32Buffer >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 11,
                       (int32Buffer >> 24) & 0xFF);

    // subsamplingFactor
    this->_blob.insert(this->_blob.begin() + viewOffset + 12,
                       (descriptor.subsamplingFactor >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 13,
                       (descriptor.subsamplingFactor >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 14,
                       (descriptor.subsamplingFactor >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 15,
                       (descriptor.subsamplingFactor >> 24) & 0xFF);

    // uriStart
    std::uint32_t uriStartOffset = gcpsOffset - stringsOffset;
    this->_blob.insert(this->_blob.begin() + viewOffset + 16,
                       (uriStartOffset >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 17,
                       (uriStartOffset >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 18,
                       (uriStartOffset >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 19,
                       (uriStartOffset >> 24) & 0xFF);

    // uriEnd
    std::uint32_t uriEndOffset {uriStartOffset + uriSize};
    this->_blob.insert(this->_blob.begin() + viewOffset + 20,
                       (uriEndOffset >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 21,
                       (uriEndOffset >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 22,
                       (uriEndOffset >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 23,
                       (uriEndOffset >> 24) & 0xFF);

    //std::cout << "Inserting URI start = " << std::to_string(uriStartOffset) << ", end = " << std::to_string(uriEndOffset) << std::endl;

    // shape xArity
    this->_blob.insert(this->_blob.begin() + viewOffset + 24,
                       (descriptor.shape.xArity >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 25,
                       (descriptor.shape.xArity >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 26,
                       (descriptor.shape.xArity >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 27,
                       (descriptor.shape.xArity >> 24) & 0xFF);

    // shape yArity
    this->_blob.insert(this->_blob.begin() + viewOffset + 28,
                       (descriptor.shape.yArity >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 29,
                       (descriptor.shape.yArity >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 30,
                       (descriptor.shape.yArity >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 31,
                       (descriptor.shape.yArity >> 24) & 0xFF);

    // gcps start
    std::uint32_t gcpsStart {0u};
    gcpsStart = originalSize - gcpsOffset;
    this->_blob.insert(this->_blob.begin() + viewOffset + 32,
                       (gcpsStart >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 33,
                       (gcpsStart >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 34,
                       (gcpsStart >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 35,
                       (gcpsStart >> 24) & 0xFF);

    // gcps end
    std::uint32_t gcpsEnd {0u};
    gcpsEnd = gcpsStart + static_cast<std::uint32_t>(descriptor.shape.gcpsLat.size()) * gcpSize;
    this->_blob.insert(this->_blob.begin() + viewOffset + 36,
                       (gcpsEnd >> 0) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 37,
                       (gcpsEnd >> 8) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 38,
                       (gcpsEnd >> 16) & 0xFF);
    this->_blob.insert(this->_blob.begin() + viewOffset + 39,
                       (gcpsEnd >> 24) & 0xFF);

    // Store uri string
    this->_blob.insert(this->_blob.begin() + gcpsOffset + descriptorSize,
                       descriptor.uri.begin(), descriptor.uri.end());

    // Store GCPs
    const std::size_t gcpsCount {descriptor.shape.gcpsLat.size()};
    std::size_t oldSize {this->_blob.size()};
    this->_blob.resize(oldSize + gcpSize * gcpsCount);
    std::uint8_t* timePtr = this->_blob.data() + oldSize;
    std::uint8_t* latPtr = timePtr + gcpsCount * sizeof(std::uint64_t);
    std::uint8_t* lonPtr = latPtr + gcpsCount * sizeof(float);
    std::uint8_t* minorAxisPtr = lonPtr + gcpsCount * sizeof(float);
    std::uint8_t* majorAxisPtr = minorAxisPtr + gcpsCount * sizeof(std::uint32_t);

    std::memcpy(timePtr, descriptor.shape.gcpsTime.data(), gcpsCount * sizeof(std::uint64_t));
    std::memcpy(latPtr, descriptor.shape.gcpsLat.data(), gcpsCount * sizeof(float));
    std::memcpy(lonPtr, descriptor.shape.gcpsLon.data(), gcpsCount * sizeof(float));
    std::memcpy(minorAxisPtr, descriptor.shape.gcpsMinorAxisIndex.data(), gcpsCount * sizeof(std::uint32_t));
    std::memcpy(majorAxisPtr, descriptor.shape.gcpsMajorAxisIndex.data(), gcpsCount * sizeof(std::uint32_t));

    // Compute new offsets and update their values in the blob
    stringsOffset = stringsOffset + descriptorSize;
    this->_blob[StringsOffsetIndex + 0] = (stringsOffset >> 0) & 0xFF;
    this->_blob[StringsOffsetIndex + 1] = (stringsOffset >> 8) & 0xFF;
    this->_blob[StringsOffsetIndex + 2] = (stringsOffset >> 16) & 0xFF;
    this->_blob[StringsOffsetIndex + 3] = (stringsOffset >> 24) & 0xFF;

    gcpsOffset = gcpsOffset + descriptorSize + uriSize;
    this->_blob[GcpsOffsetIndex + 0] = (gcpsOffset >> 0) & 0xFF;
    this->_blob[GcpsOffsetIndex + 1] = (gcpsOffset >> 8) & 0xFF;
    this->_blob[GcpsOffsetIndex + 2] = (gcpsOffset >> 16) & 0xFF;
    this->_blob[GcpsOffsetIndex + 3] = (gcpsOffset >> 24) & 0xFF;
}

/**
 *
 */
std::uint32_t GranuleMetadata::sourceId() const
{
    return ( (static_cast<std::uint32_t>(this->_blob[SourceIdIndex + 0]) << 0)
           | (static_cast<std::uint32_t>(this->_blob[SourceIdIndex + 1]) << 8)
           | (static_cast<std::uint32_t>(this->_blob[SourceIdIndex + 2]) << 16)
           | (static_cast<std::uint32_t>(this->_blob[SourceIdIndex + 3]) << 24));
}

/**
 *
 */
std::uint32_t GranuleMetadata::collectionId() const
{
    return ( (static_cast<std::uint32_t>(this->_blob[CollectionIdIndex + 0]) << 0)
           | (static_cast<std::uint32_t>(this->_blob[CollectionIdIndex + 1]) << 8)
           | (static_cast<std::uint32_t>(this->_blob[CollectionIdIndex + 2]) << 16)
           | (static_cast<std::uint32_t>(this->_blob[CollectionIdIndex + 3]) << 24));
}

/**
 *
 */
std::uint64_t GranuleMetadata::granuleId() const
{
    return ( (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 0]) << 0)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 1]) << 8)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 2]) << 16)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 3]) << 24)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 4]) << 32)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 5]) << 40)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 6]) << 48)
           | (static_cast<std::uint64_t>(this->_blob[GranuleIdIndex + 7]) << 56));
}

/**
 *
 */
DataModel GranuleMetadata::dataModel() const
{
    return static_cast<DataModel>(this->_blob[DataModelIndex]);
}

/**
 *
 */
timestamp_t GranuleMetadata::timeCoverageStart() const
{
    return ( (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 0]) << 0)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 1]) << 8)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 2]) << 16)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 3]) << 24)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 4]) << 32)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 5]) << 40)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 6]) << 48)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageStartIndex + 7]) << 56));
}

/**
 *
 */
timestamp_t GranuleMetadata::timeCoverageEnd() const
{
    return ( (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 0]) << 0)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 1]) << 8)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 2]) << 16)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 3]) << 24)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 4]) << 32)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 5]) << 40)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 6]) << 48)
           | (static_cast<std::uint64_t>(this->_blob[TimeCoverageEndIndex + 7]) << 56));
}

/**
 *
 */
float GranuleMetadata::latMin() const
{
    float value {0.f};  //
    int32_t int32Buffer {0};    //

    int32Buffer = ( (static_cast<std::int32_t>(this->_blob[LatMinIndex + 0]) << 0)
                  | (static_cast<std::int32_t>(this->_blob[LatMinIndex + 1]) << 8)
                  | (static_cast<std::int32_t>(this->_blob[LatMinIndex + 2]) << 16)
                  | (static_cast<std::int32_t>(this->_blob[LatMinIndex + 3]) << 24));

    memcpy(&value, &int32Buffer, 4);

    return value;
}

/**
 *
 */
float GranuleMetadata::latMax() const
{
    float value {0.f};  //
    int32_t int32Buffer {0};    //

    int32Buffer = ( (static_cast<std::int32_t>(this->_blob[LatMaxIndex + 0]) << 0)
                  | (static_cast<std::int32_t>(this->_blob[LatMaxIndex + 1]) << 8)
                  | (static_cast<std::int32_t>(this->_blob[LatMaxIndex + 2]) << 16)
                  | (static_cast<std::int32_t>(this->_blob[LatMaxIndex + 3]) << 24));

    memcpy(&value, &int32Buffer, 4);

    return value;
}

/**
 *
 */
float GranuleMetadata::lonMin() const
{
    float value {0.f};  //
    int32_t int32Buffer {0};    //

    int32Buffer = ( (static_cast<std::int32_t>(this->_blob[LonMinIndex + 0]) << 0)
                  | (static_cast<std::int32_t>(this->_blob[LonMinIndex + 1]) << 8)
                  | (static_cast<std::int32_t>(this->_blob[LonMinIndex + 2]) << 16)
                  | (static_cast<std::int32_t>(this->_blob[LonMinIndex + 3]) << 24));

    memcpy(&value, &int32Buffer, 4);

    return value;
}

/**
 *
 */
float GranuleMetadata::lonMax() const
{
    float value {0.f};  //
    int32_t int32Buffer {0};    //

    int32Buffer = ( (static_cast<std::int32_t>(this->_blob[LonMaxIndex + 0]) << 0)
                  | (static_cast<std::int32_t>(this->_blob[LonMaxIndex + 1]) << 8)
                  | (static_cast<std::int32_t>(this->_blob[LonMaxIndex + 2]) << 16)
                  | (static_cast<std::int32_t>(this->_blob[LonMaxIndex + 3]) << 24));

    memcpy(&value, &int32Buffer, 4);

    return value;
}

/**
 *
 */
std::string GranuleMetadata::dataId() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get dataIdEnd
    std::uint32_t dataIdEnd {0u};
    dataIdEnd = ( (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 0]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 1]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 2]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 3]) << 24));

    //std::cout << "dataIdEnd" << std::to_string(dataIdEnd) << std::endl;
    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset, dataIdEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::title() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get dataIdEnd
    std::uint32_t dataIdEnd {0u};
    dataIdEnd = ( (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 0]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 1]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 2]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[DataIdIndex + 3]) << 24));

    // Get titleEnd
    std::uint32_t titleEnd {0u};
    titleEnd = ( (static_cast<std::uint32_t>(this->_blob[TitleIndex + 0]) << 0)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 1]) << 8)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 2]) << 16)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 3]) << 24));
    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + dataIdEnd, titleEnd - dataIdEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::institution() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get titleEnd
    std::uint32_t titleEnd {0u};
    titleEnd = ( (static_cast<std::uint32_t>(this->_blob[TitleIndex + 0]) << 0)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 1]) << 8)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 2]) << 16)
               | (static_cast<std::uint32_t>(this->_blob[TitleIndex + 3]) << 24));

    // Get institutionEnd
    std::uint32_t institutionEnd {0u};
    institutionEnd = ( (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 0]) << 0)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 1]) << 8)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 2]) << 16)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + titleEnd, institutionEnd - titleEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::comment() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get institutionEnd
    std::uint32_t institutionEnd {0u};
    institutionEnd = ( (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 0]) << 0)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 1]) << 8)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 2]) << 16)
                     | (static_cast<std::uint32_t>(this->_blob[InstitutionIndex + 3]) << 24));

    // Get commentEnd
    std::uint32_t commentEnd {0u};
    commentEnd = ( (static_cast<std::uint32_t>(this->_blob[CommentIndex + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + institutionEnd, commentEnd - institutionEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::fileId() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get commentEnd
    std::uint32_t commentEnd {0u};
    commentEnd = ( (static_cast<std::uint32_t>(this->_blob[CommentIndex + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[CommentIndex + 3]) << 24));

    // Get fileIdEnd
    std::uint32_t fileIdEnd {0u};
    fileIdEnd = ( (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 0]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 1]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 2]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + commentEnd, fileIdEnd - commentEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::productVersion() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get fileIdEnd
    std::uint32_t fileIdEnd {0u};
    fileIdEnd = ( (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 0]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 1]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 2]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[FileIdIndex + 3]) << 24));

    // Get productVersionEnd
    std::uint32_t productVersionEnd {0u};
    productVersionEnd = ( (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 0]) << 0)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 1]) << 8)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 2]) << 16)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + fileIdEnd, productVersionEnd - fileIdEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::creatorEmail() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get productVersionEnd
    std::uint32_t productVersionEnd {0u};
    productVersionEnd = ( (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 0]) << 0)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 1]) << 8)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 2]) << 16)
                        | (static_cast<std::uint32_t>(this->_blob[ProductVersionIndex + 3]) << 24));

    // Get creatorEmailEnd
    std::uint32_t creatorEmailEnd {0u};
    creatorEmailEnd = ( (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 0]) << 0)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 1]) << 8)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 2]) << 16)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + productVersionEnd, creatorEmailEnd - productVersionEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::stationId() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get creatorEmailEnd
    std::uint32_t creatorEmailEnd {0u};
    creatorEmailEnd = ( (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 0]) << 0)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 1]) << 8)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 2]) << 16)
                      | (static_cast<std::uint32_t>(this->_blob[CreatorEmailIndex + 3]) << 24));

    // Get stationIdEnd
    std::uint32_t stationIdEnd {0u};
    stationIdEnd = ( (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 0]) << 0)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 1]) << 8)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 2]) << 16)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + creatorEmailEnd, stationIdEnd - creatorEmailEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::platform() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get stationIdEnd
    std::uint32_t stationIdEnd {0u};
    stationIdEnd = ( (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 0]) << 0)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 1]) << 8)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 2]) << 16)
                   | (static_cast<std::uint32_t>(this->_blob[StationIdIndex + 3]) << 24));

    // Get platformEnd
    std::uint32_t platformEnd {0u};
    platformEnd = ( (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 0]) << 0)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 1]) << 8)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 2]) << 16)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + stationIdEnd, platformEnd - stationIdEnd};
    return result;
}

/**
 *
 */
std::string GranuleMetadata::sensor() const
{
    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get platformEnd
    std::uint32_t platformEnd {0u};
    platformEnd = ( (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 0]) << 0)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 1]) << 8)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 2]) << 16)
                  | (static_cast<std::uint32_t>(this->_blob[PlatformIndex + 3]) << 24));

    // Get sensorEnd
    std::uint32_t sensorEnd {0u};
    sensorEnd = ( (static_cast<std::uint32_t>(this->_blob[SensorIndex + 0]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[SensorIndex + 1]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[SensorIndex + 2]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[SensorIndex + 3]) << 24));

    std::string result {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + platformEnd, sensorEnd - platformEnd};
    return result;
}

/**
 *
 */
std::uint32_t GranuleMetadata::viewsCount() const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);

    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    std::uint32_t count = (stringsOffset - ViewsOffsetIndex) / descriptorSize;
    return count;
}

/**
 *
 */
bool GranuleMetadata::getView(std::uint32_t index, IDFDescriptor& idf) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    constexpr std::uint32_t gcpSize = sizeof(std::uint64_t) + 2 * sizeof(std::uint32_t) + 2 * sizeof(float);
    const std::uint32_t descriptorOffset {ViewsOffsetIndex + index * descriptorSize};
    std::int32_t int32Buffer {0};   //
    std::uint32_t startOffset {0u};    //
    std::uint32_t endOffset {0u};   //

    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    // Get current gcpsOffset
    std::uint32_t gcpsOffset {0u};
    gcpsOffset = ( (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[GcpsOffsetIndex + 3]) << 24));

    idf.xArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 3]) << 24));

    idf.yArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 4]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 5]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 6]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 7]) << 24));

    int32Buffer = ( (static_cast<std::int32_t>(this->_blob[descriptorOffset + 8]) << 0)
                  | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 9]) << 8)
                  | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 10]) << 16)
                  | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 11]) << 24));
    memcpy(&(idf.resolution), &int32Buffer, 4);

    idf.subsamplingFactor = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 12]) << 0)
                            | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 13]) << 8)
                            | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 14]) << 16)
                            | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 15]) << 24));

    startOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 16]) << 0)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 17]) << 8)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 18]) << 16)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 19]) << 24));

    endOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 20]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 21]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 22]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 23]) << 24));

    //std::cout << "getting " << std::to_string(index) << "th view" << std::endl;
    //std::cout << "stringsOffset = " << std::to_string(stringsOffset) << ", startOffset = " << std::to_string(startOffset) << ", endOffset = " << std::to_string(endOffset) << std::endl;
    std::string uri {reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + startOffset, endOffset - startOffset};
    idf.uri = uri;  // probably possible to avoid copy


    idf.shape.xArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 24]) << 0)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 25]) << 8)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 26]) << 16)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 27]) << 24));

    idf.shape.yArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 28]) << 0)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 29]) << 8)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 30]) << 16)
                       | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 31]) << 24));

    startOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 32]) << 0)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 33]) << 8)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 34]) << 16)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 35]) << 24));

    endOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 36]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 37]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 38]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 39]) << 24));

    //std::cout << "shape.xArity = " << std::to_string(idf.shape.xArity) << ", shape.yArity = " << std::to_string(idf.shape.yArity) << ", uri = " << idf.uri << std::endl;
    //std::cout << "gcpsOffset = " << std::to_string(gcpsOffset) << ", startOffset = " << std::to_string(startOffset) << ", endOffset = " << std::to_string(endOffset) <<std::endl;
    std::size_t gcpsCount {0u}; //
    gcpsCount = static_cast<std::size_t>((endOffset - startOffset) / gcpSize);
    idf.shape.gcpsTime.resize(gcpsCount);
    idf.shape.gcpsLat.resize(gcpsCount);
    idf.shape.gcpsLon.resize(gcpsCount);
    idf.shape.gcpsMinorAxisIndex.resize(gcpsCount);
    idf.shape.gcpsMajorAxisIndex.resize(gcpsCount);

    const std::uint8_t* timePtr = this->_blob.data() + gcpsOffset + startOffset;
    const std::uint8_t* latPtr = timePtr + gcpsCount * sizeof(std::uint64_t);
    const std::uint8_t* lonPtr = latPtr + gcpsCount * sizeof(float);
    const std::uint8_t* minorAxisPtr = lonPtr + gcpsCount * sizeof(float);
    const std::uint8_t* majorAxisPtr = minorAxisPtr + gcpsCount * sizeof(std::uint32_t);

    std::memcpy(idf.shape.gcpsTime.data(), timePtr, gcpsCount * sizeof(std::uint64_t));
    std::memcpy(idf.shape.gcpsLat.data(), latPtr, gcpsCount * sizeof(float));
    std::memcpy(idf.shape.gcpsLon.data(), lonPtr, gcpsCount * sizeof(float));
    std::memcpy(idf.shape.gcpsMinorAxisIndex.data(), minorAxisPtr, gcpsCount * sizeof(std::uint32_t));
    std::memcpy(idf.shape.gcpsMajorAxisIndex.data(), majorAxisPtr, gcpsCount * sizeof(std::uint32_t));

    return true;
}

/**
 *
 */
bool GranuleMetadata::getViewURI(std::uint32_t index, std::string& result) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
    const std::uint32_t descriptorOffset {ViewsOffsetIndex + index * descriptorSize};
    std::uint32_t startOffset {0u};    //
    std::uint32_t endOffset {0u};   //

    // Get current stringsOffset
    std::uint32_t stringsOffset {0u};
    stringsOffset = ( (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 0]) << 0)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 1]) << 8)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 2]) << 16)
                    | (static_cast<std::uint32_t>(this->_blob[StringsOffsetIndex + 3]) << 24));

    startOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 16]) << 0)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 17]) << 8)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 18]) << 16)
                  | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 19]) << 24));

    endOffset = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 20]) << 0)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 21]) << 8)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 22]) << 16)
                | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 23]) << 24));

    result = std::string{reinterpret_cast<const char*>(this->_blob.data()) + stringsOffset + startOffset, endOffset - startOffset};
    return true;
}

/**
 *
 */
bool GranuleMetadata::selectResolution(std::uint32_t viewportPixelArea,
                                           float viewportSpacingInMeters,
                                           float minScreenEstate,
                                           std::uint32_t maxTextureSize,
                                           IDFDescriptor& idf) const
{
    constexpr std::uint32_t descriptorSize = 9 * sizeof(std::uint32_t) + sizeof(float);
	std::int64_t j {0};	//
	float deltaRes {0.f};	//
    float viewportSize {0.f};   //
	float granuleSize {0.f};	//
	unsigned int defaultIndex {0u}; //
    std::uint32_t xArity {0u};  //
    std::uint32_t yArity {0u};  //
    std::int32_t int32Buffer {0};   //
    float spacingInMeters {0.f};    //

    std::size_t viewsCount {this->viewsCount()};    //

    viewportSize = static_cast<float>(static_cast<double>(viewportSpacingInMeters) * std::sqrt(viewportPixelArea));
	j = static_cast<std::int64_t>(viewsCount) - 1;
	for
		(; j > -1; --j)
	{
        std::uint32_t k = static_cast<std::uint32_t>(j);
        std::uint32_t descriptorOffset = ViewsOffsetIndex + k * descriptorSize;

        xArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 3]) << 24));

        yArity = ( (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 4]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 5]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 6]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[descriptorOffset + 7]) << 24));

		/* Temporary workaround for high resolution images */
		if
			(  xArity > maxTextureSize
			|| yArity > maxTextureSize)
		{
			Log::log(LogLevel::Error, "GranuleMetadata", "A resolution was skipped because the GPU would not have been able to handle it");
			defaultIndex += 1;
			continue;
		}

        int32Buffer = ( (static_cast<std::int32_t>(this->_blob[descriptorOffset + 8]) << 0)
                      | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 9]) << 8)
                      | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 10]) << 16)
                      | (static_cast<std::int32_t>(this->_blob[descriptorOffset + 11]) << 24));
        memcpy(&spacingInMeters, &int32Buffer, 4);

		deltaRes = viewportSpacingInMeters - spacingInMeters;
		Log::log(LogLevel::Debug, "GranuleMetadata", "Check res {}: 0 < {} ? {}",
                 spacingInMeters, deltaRes, 0.f < deltaRes);
		if
			(0.f < deltaRes)
		{
			granuleSize = static_cast<float>(static_cast<double>(spacingInMeters) * std::sqrt(xArity * xArity + yArity * yArity));
			Log::log(LogLevel::Debug, "GranuleMetadata", "Screen estate: {}m / {}m => {}"
					, granuleSize, viewportSize, minScreenEstate);
			if
				(minScreenEstate > granuleSize / viewportSize)
			{
				// The portion of the viewport occupied by the granule is below
                // the threshold
				if
					(k < viewsCount - 1)
				{
					// Select lower resolution since the higher resolution
					// is too far away from the viewport res
                    return this->getView(k + 1, idf);
				}

				// no lower resolution, higher resolution too far
				return false;
			}

			// found a suitable higher res
            return this->getView(k, idf);
		}

		// Ignore lower res data
	}

	// No suitable higher res data found, use closest low res
	Log::log(LogLevel::Debug, "GranuleMetadata", "Use first resolution by default");
	if
		(defaultIndex >= viewsCount)
	{
		Log::log(LogLevel::Error, "GranuleMetadata",
                 "Even first resolution is not suitable (probably a dimension exceeding GPU memory capacity)");
		return false;
	}
    return this->getView(defaultIndex, idf);
};


/**
 *
 */
bool GranuleMetadata::has(IDFAttributes attr) const
{
    return 0 < ( ( (static_cast<std::uint32_t>(this->_blob[AvailableAttributesIndex + 0]) << 0)
                 | (static_cast<std::uint32_t>(this->_blob[AvailableAttributesIndex + 1]) << 8)
                 | (static_cast<std::uint32_t>(this->_blob[AvailableAttributesIndex + 2]) << 16)
                 | (static_cast<std::uint32_t>(this->_blob[AvailableAttributesIndex + 3]) << 24)
                 )
               & (1u << static_cast<std::size_t>(attr))
               );
}

/**
 *
 */
void GranuleMetadata::sourceId(std::uint32_t value)
{
    this->_blob[SourceIdIndex + 0] = (value >> 0) & 0xFF;
    this->_blob[SourceIdIndex + 1] = (value >> 8) & 0xFF;
    this->_blob[SourceIdIndex + 2] = (value >> 16) & 0xFF;
    this->_blob[SourceIdIndex + 3] = (value >> 24) & 0xFF;
}

/**
 *
 */
void GranuleMetadata::collectionId(std::uint32_t value)
{
    this->_blob[CollectionIdIndex + 0] = (value >> 0) & 0xFF;
    this->_blob[CollectionIdIndex + 1] = (value >> 8) & 0xFF;
    this->_blob[CollectionIdIndex + 2] = (value >> 16) & 0xFF;
    this->_blob[CollectionIdIndex + 3] = (value >> 24) & 0xFF;
}

/**
 *
 */
void GranuleMetadata::granuleId(std::uint64_t value)
{
    this->_blob[GranuleIdIndex + 0] = (value >> 0) & 0xFF;
    this->_blob[GranuleIdIndex + 1] = (value >> 8) & 0xFF;
    this->_blob[GranuleIdIndex + 2] = (value >> 16) & 0xFF;
    this->_blob[GranuleIdIndex + 3] = (value >> 24) & 0xFF;
    this->_blob[GranuleIdIndex + 4] = (value >> 32) & 0xFF;
    this->_blob[GranuleIdIndex + 5] = (value >> 40) & 0xFF;
    this->_blob[GranuleIdIndex + 6] = (value >> 48) & 0xFF;
    this->_blob[GranuleIdIndex + 7] = (value >> 56) & 0xFF;
}

/**
 *
 */
void GranuleMetadata::timeCoverageStart(timestamp_t value)
{
   this->_blob[TimeCoverageStartIndex + 0] = (value >> 0) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 1] = (value >> 8) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 2] = (value >> 16) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 3] = (value >> 24) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 4] = (value >> 32) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 5] = (value >> 40) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 6] = (value >> 48) & 0xFF;
   this->_blob[TimeCoverageStartIndex + 7] = (value >> 56) & 0xFF;
}

/**
 *
 */
void GranuleMetadata::timeCoverageEnd(timestamp_t value)
{
    this->_blob[TimeCoverageEndIndex + 0] = (value >> 0) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 1] = (value >> 8) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 2] = (value >> 16) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 3] = (value >> 24) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 4] = (value >> 32) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 5] = (value >> 40) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 6] = (value >> 48) & 0xFF;
    this->_blob[TimeCoverageEndIndex + 7] = (value >> 56) & 0xFF;
}

/**
 *
 */
const std::uint8_t* GranuleMetadata::getBlob(std::uint64_t& size) const
{
    size = this->_blob.size();
    return this->_blob.data();
}

/**
 *
 */
void GranuleMetadata::setBlob(std::uint8_t* value, std::uint64_t size)
{
    this->_blob.assign(value, value + size);

    //std::cout << "setBlob(0x???, " << std::to_string(size) << "):" << std::endl;
    //std::cout << "\tsourceId: " << std::to_string(this->sourceId()) << std::endl;
    //std::cout << "\tcollectionId: " << std::to_string(this->collectionId()) << std::endl;
    //std::cout << "\tgranuleId: " << std::to_string(this->granuleId()) << std::endl;
    //std::cout << "\tdataModel: " << std::to_string(static_cast<int>(this->dataModel())) << std::endl;
    //std::cout << "\ttimeCoverageStart: " << std::to_string(this->timeCoverageStart()) << std::endl;
    //std::cout << "\ttimeCoverageEnd: " << std::to_string(this->timeCoverageEnd()) << std::endl;
    //std::cout << "\tlatMin: " << std::to_string(this->latMin()) << std::endl;
    //std::cout << "\tlatMax: " << std::to_string(this->latMax()) << std::endl;
    //std::cout << "\tlonMin: " << std::to_string(this->lonMin()) << std::endl;
    //std::cout << "\tlonMax: " << std::to_string(this->lonMax()) << std::endl;
    //std::cout << "\tdataId: " << this->dataId() << std::endl;
    //std::cout << "\ttitle: " << this->title() << std::endl;
    //std::cout << "\tinstitution: " << this->institution() << std::endl;
    //std::cout << "\tcomment: " << this->comment() << std::endl;
    //std::cout << "\tfileId: " << this->fileId() << std::endl;
    //std::cout << "\tproductVersion: " << this->productVersion() << std::endl;
    //std::cout << "\tstationId: " << this->stationId() << std::endl;
    //std::cout << "\tplatform: " << this->platform() << std::endl;
    //std::cout << "\tsensor: " << this->sensor() << std::endl;
    //std::cout << "\tviewsCount: " << std::to_string(this->viewsCount()) << std::endl;
}
