// This file is part of SEAScope, a 3D visualisation and analysis application
// for satellite, in-situ and numerical model data.
//
// Copyright (C) 2014-2023 OceanDataLab
//
// SEAScope is free software: you can redistribute it and/or modify it under
// the terms of the GNU Lesser General Public License as published by the Free
// Software Foundation, either version 3 of the License, or (at your option)
// any later version.
//
// SEAScope is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
// details.
//
// You should have received a copy of the GNU Lesser General Public License
// along with SEAScope. If not, see <https://www.gnu.org/licenses/>.
#include <ctime>
#include <algorithm>
#include <unordered_map>
#include "vendors/_netcdf.h"

#include "local/netcdf_utils.h"
#include "core/structs.h"
#include "core/log.h"
#include "core/utils.h"
#include "core/granule_metadata.h"

#ifdef _WIN32
	#include "core/strptime.h"
#endif


/**
 *
 */
netCDF::NcVarAtt SEAScope::readAttribute(netCDF::NcVar& var, const std::string& attrName, bool& found)
{
	netCDF::NcVarAtt attribute;	//
	auto attrMap = var.getAtts();
	auto it = attrMap.find(attrName);

	if
		(it != attrMap.end())
	{
		found = true;
		return it->second;
	}
	found = false;
	return attribute;
};

/**
 *
 */
bool SEAScope::str2timestamp(const std::string& timeStr, std::time_t& result)
{
	std::size_t microsecondsPosition;	//
	std::int32_t microseconds;	//
	std::string withoutMicro;	//
	std::string microsecondsStr;	//
	std::tm t = {};	//
	std::time_t timestamp;	//

	microseconds = 0;
	microsecondsPosition = timeStr.find_first_of(".");
	withoutMicro = timeStr.substr(0, microsecondsPosition);
	if
		(std::string::npos != microsecondsPosition)
	{
		microsecondsStr = timeStr.substr( 1 + microsecondsPosition
										, timeStr.length() - 1);
		microseconds = std::atoi(microsecondsStr.c_str());
	}

    // Try to parse with multiple formats
    std::array<std::string, 8> timeFormats = { {
        "%Y-%m-%dT%H:%M:%S",
        "%Y%m%dT%H%M%S",
        "%Y%m%d %H%M%S",
        "%Y-%m-%d %H:%M:%S",
        "%Y%m%d%H%M%S",
        "%Y-%m-%d",
        "%Y%m%d",
        "%Y-%j"
    } };

    bool success {false};
    char* parsed {nullptr};
    for
        (const std::string& timeFormat : timeFormats)
    {
        parsed = strptime(withoutMicro.c_str(), timeFormat.c_str(), &t);
        if
            (nullptr == parsed)
        {
            continue;
        }

        success = true;
        break;
    }
    parsed = nullptr;

    if
        (false == success)
    {
        Log::log(LogLevel::Error, "NetCDF Utils",
                 "Could not parse '{}' with any of the supported datetime formats", timeStr);
        return false;
    }

	timestamp = std::mktime(&t);
	/* result in milliseconds */
	result = 1000 * timestamp + static_cast<time_t>(0.001 * static_cast<double>(microseconds));
	Log::log( LogLevel::Debug, "IDFReader"
			, "Time coverage extremum: {} {} ### 1000 * {} + .001 * {} = {}"
			, timeStr, withoutMicro, timestamp, microseconds, result);
    return true;
};


/**
 *
 */
bool SEAScope::extractOptionalAttributes(netCDF::NcFile& fileHandler,
                                         IDFOptionalAttributes& info)
{
    netCDF::NcGroupAtt attr;

	// title
	attr = fileHandler.getAtt("title");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.title);
	}

	// institution
	attr = fileHandler.getAtt("institution");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.institution);
	}

	// comment
	attr = fileHandler.getAtt("comment");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.comment);
	}

	// [product_]id
	attr = fileHandler.getAtt("id");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.fileId);
	}

	// product_version
	attr = fileHandler.getAtt("product_version");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.productVersion);
	}

	// geospatial_lat_min
	attr = fileHandler.getAtt("geospatial_lat_min");
	if
		(false == attr.isNull())
	{
        info.hasLatMin = true;
		attr.getValues(&info.latMin);
	}

	// geospatial_lat_max
	attr = fileHandler.getAtt("geospatial_lat_max");
	if
		(false == attr.isNull())
	{
        info.hasLatMax = true;
		attr.getValues(&info.latMax);
	}

	// geospatial_lon_min
	attr = fileHandler.getAtt("geospatial_lon_min");
	if
		(false == attr.isNull())
	{
        info.hasLonMin = true;
		attr.getValues(&info.lonMin);
	}

	// geospatial_lon_max
	attr = fileHandler.getAtt("geospatial_lon_max");
	if
		(false == attr.isNull())
	{
        info.hasLonMax = true;
		attr.getValues(&info.lonMax);
	}

	// creator_email
	attr = fileHandler.getAtt("creator_email");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.creatorEmail);
	}

	// station_id
	attr = fileHandler.getAtt("station_id");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.stationId);
	}

	// platform
	attr = fileHandler.getAtt("platform");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.platform);
	}

	// sensor
	attr = fileHandler.getAtt("sensor");
	if
		(false == attr.isNull())
	{
		attr.getValues(info.sensor);
	}

    return true;
}

/**
 *
 */
bool SEAScope::parseTimeUnits(const std::string& timeStr, std::time_t& epoch, double& multiplier)
{
    std::vector<std::string> tokens;
    SEAScope::explodeString(timeStr, " ", tokens);

    if
        (3 > tokens.size())
    {
        Log::log(LogLevel::Error, "NetCDF Utils",
                 "Time units string does not match the expected format: 'UNITS since EPOCH'");
        return false;
    }

    const std::unordered_map<std::string, double> unitsToMilliseconds {
        /* microseconds */
        {"microseconds", 0.001},
        {"microsecond", 0.001},
        {"microsecs", 0.001},
        {"microsec", 0.001},

        /* milliseconds */
        {"milliseconds", 1.0},
        {"millisecond", 1.0},
        {"millisecs", 1.0},
        {"millisec", 1.0},
        {"msecs", 1.0},
        {"msec", 1.0},
        {"ms", 1.0},

        /* seconds */
        {"seconds", 1000.0},
        {"second", 1000.0},
        {"secs", 1000.0},
        {"sec", 1000.0},
        {"s", 1000.0},

        /* minutes */
        {"minutes", 60 * 1000.0},
        {"minute", 60 * 1000.0},
        {"mins", 60 * 1000.0},
        {"min", 60 * 1000.0},

        /* hours */
        {"hour", 3600 * 1000.0},
        {"hours", 3600 * 1000.0},
        {"hr", 3600 * 1000.0},
        {"hrs", 3600 * 1000.0},
        {"h", 3600 * 1000.0},

        /* days */
        {"day", 24 * 3600 * 1000.0},
        {"days", 24 * 3600 * 1000.0},
        {"d", 24 * 3600 * 1000.0}
    };

    std::string lowerUnitsStr;
    lowerUnitsStr.resize(tokens[0].size());
    std::transform(tokens[0].cbegin(),
                   tokens[0].cend(),
                   lowerUnitsStr.begin(),
                   [] (unsigned char c) {return std::tolower(c);});
    auto unitsFound = unitsToMilliseconds.find(lowerUnitsStr);
    if
        (unitsToMilliseconds.end() == unitsFound)
    {
        Log::log(LogLevel::Error, "NetCDF Utils", "Unknown time units '{}'", lowerUnitsStr);
        return false;
    }
    multiplier = unitsFound->second;

    tokens.erase(tokens.begin(), std::next(std::next(tokens.begin())));

    std::string epochStr = fmt::format("{}", fmt::join(tokens, " "));
    SEAScope::str2timestamp(epochStr, epoch);

    return true;
}

/**
 *
 */
bool SEAScope::getDepthIndex(const netCDF::NcFile& fileHandler, const std::string& depthName, float depthValue, std::size_t& result)
{
    netCDF::NcVar depthVar;

    depthVar = fileHandler.getVar(depthName);
    if
        (true == depthVar.isNull())
    {
        Log::log(LogLevel::Error, "NetCDF Utils", "Depth variable '{}' not found", depthName);
        return false;
    }

    if
        (1 != depthVar.getDimCount())
    {
        Log::log(LogLevel::Error, " NetCDF Utils",
                 "Depth variable '{}' does not have exactly one dimension, it is not supported", depthName);
        return false;
    }

    // Read all depth values
    float* depthValues {nullptr};
    depthValues = new float[depthVar.getDim(0).getSize()];
    depthVar.getVar(depthValues);

    std::vector<float> depthVector(depthValues, depthValues + depthVar.getDim(0).getSize());
    auto depthIter = std::lower_bound(depthVector.cbegin(), depthVector.cend(),depthValue);
    long index = std::distance(depthVector.cbegin(), depthIter);
    float nextValue = *depthIter;
    float prevValue = *std::prev(depthIter);
    if
        (std::abs(nextValue - depthValue) > std::abs(prevValue - depthValue))
    {
        result = static_cast<std::size_t>(std::max({0L, index - 1}));
    }
    else
    {
        result = static_cast<std::size_t>(std::min({index, static_cast<long>(depthVar.getDim(0).getSize() - 1)}));
    }

    delete[] depthValues;
    return true;
}
